#include <thread>
#include <mutex>
#include <condition_variable>
#include <stdlib.h>
#include <queue>
#include <iostream>

template<typename Data>
class concurrent_queue
{
    std::mutex the_mutex;
    std::queue<Data> the_queue;
    std::condition_variable the_cv;
public:
    void push(const Data& data)
    {
        {
            std::lock_guard<std::mutex> lk(the_mutex);
            the_queue.push(data);
        }
        the_cv.notify_one();
    }

    bool try_pop(Data& data)
    {
        std::lock_guard<std::mutex> lk(the_mutex);
        if(the_queue.empty()) return false;
        data=the_queue.front();
        the_queue.pop();
        return true;
    }

    void wait_and_pop(Data& data)
    {
        std::unique_lock<std::mutex> lk(the_mutex);
        while(the_queue.empty())
        {
            the_cv.wait(lk);
        }
        data=the_queue.front();
        the_queue.pop();
    }
};

void pusher(concurrent_queue<int>* q,unsigned long count)
{
    for(unsigned long i=0;i<count;++i)
    {
        q->push(rand());
    }
}

void popper(concurrent_queue<int>* q,unsigned long count)
{
    for(unsigned long i=0;i<count;++i)
    {
        int x;
        q->wait_and_pop(x);
    }
}

int main()
{
    concurrent_queue<int> q;
    unsigned long const count=10000000;

    std::chrono::monotonic_clock::time_point start=std::chrono::monotonic_clock::now();
    pusher(&q,count);
    popper(&q,count);
    std::chrono::monotonic_clock::time_point stop=std::chrono::monotonic_clock::now();

    std::cout<<"Time taken (one thread)="<<std::chrono::duration_cast<std::chrono::milliseconds>(stop-start).count()<<"ms"<<std::endl;
    
    start=std::chrono::monotonic_clock::now();
    std::thread push_thread(pusher,&q,count);
    std::thread pop_thread(popper,&q,count);

    push_thread.join();
    pop_thread.join();

    stop=std::chrono::monotonic_clock::now();
    std::cout<<"Time taken (two threads)="<<std::chrono::duration_cast<std::chrono::milliseconds>(stop-start).count()<<"ms"<<std::endl;
    
}
